//
//  XLDMusicBrainzRelease.m
//  XLD
//
//  Created by tmkk on 11/05/28.
//  Copyright 2011 tmkk. All rights reserved.
//

#import "XLDMusicBrainzRelease.h"
#import "XLDCustomClasses.h"

@implementation XLDMusicBrainzRelease

- (id)initWithReleaseID:(NSString *)releaseid andDiscID:(NSString *)discid
{
	self = [super init];
	if(!self) return nil;
	
	if(!releaseid || !discid) {
		[super dealloc];
		return nil;
	}
	
	NSURL *url = [NSURL URLWithString:[NSString stringWithFormat:@"http://musicbrainz.org/ws/2/release/%@?inc=artist-credits+recordings+discids",releaseid]];
	NSData *data = [NSData fastDataWithContentsOfURL:url];
	if(!data || [data length] == 0) {
		[super dealloc];
		return nil;
	}
	NSXMLDocument *xml = [[NSXMLDocument alloc] initWithData:data options:NSXMLNodePreserveWhitespace error:nil];
	if(!xml) {
		[super dealloc];
		return nil;
	}
	
	release = [[NSMutableDictionary alloc] init];
	NSArray *arr = [xml nodesForXPath:@"/metadata/release" error:nil];
	if(![arr count]) {
		[release release];
		[super dealloc];
		return nil;
	}
	
	id rel = [arr objectAtIndex:0];
	NSArray *objs = [rel nodesForXPath:@"./title" error:nil];
	if([objs count]) [release setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Title"];
	objs = [rel nodesForXPath:@"./artist-credit/name-credit/artist/name" error:nil];
	if([objs count] == 1) [release setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Artist"];
	else {
		NSArray *credits = [rel nodesForXPath:@"./artist-credit/name-credit" error:nil];
		NSMutableString *str = [NSMutableString string];
		int j;
		for(j=0;j<[credits count];j++) {
			id node = [credits objectAtIndex:j];
			NSString *artist = nil;
			NSString *joinphrase = nil;
			objs = [node nodesForXPath:@"./@joinphrase" error:nil];
			if([objs count]) joinphrase = [[objs objectAtIndex:0] stringValue];
			objs = [node nodesForXPath:@"./artist/name" error:nil];
			if([objs count]) artist = [[objs objectAtIndex:0] stringValue];
			if(artist && joinphrase) [str appendFormat:@"%@%@",artist,joinphrase];
			else if(artist) [str appendFormat:@"%@",artist];
		}
		[release setObject:str forKey:@"Artist"];
	}
	objs = [rel nodesForXPath:@"./date" error:nil];
	if([objs count]) [release setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Date"];
	objs = [rel nodesForXPath:@"./asin" error:nil];
	if([objs count]) [release setObject:[[objs objectAtIndex:0] stringValue] forKey:@"ASIN"];
	objs = [rel nodesForXPath:@"./barcode" error:nil];
	if([objs count]) [release setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Barcode"];
	
	NSArray *discs = [rel nodesForXPath:@"./medium-list/medium" error:nil];
	int i,j;
	for(i=0;i<[discs count];i++) {
		id disc = [discs objectAtIndex:i];
		objs = [disc nodesForXPath:@"./disc-list/disc/@id" error:nil];
		if(![objs count]) continue;
		if(![[[objs objectAtIndex:0] stringValue] isEqualToString:discid]) continue;
		objs = [disc nodesForXPath:@"./title" error:nil];
		if([objs count]) [release setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Title"];
		
		NSArray *tracks = [disc nodesForXPath:@"./track-list/track" error:nil];
		NSMutableDictionary *trackList = [NSMutableDictionary dictionary];
		for(j=0;j<[tracks count];j++) {
			id tr = [tracks objectAtIndex:j];
			NSMutableDictionary *track = [NSMutableDictionary dictionary];
			objs = [tr nodesForXPath:@"./recording/title" error:nil];
			if([objs count]) [track setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Title"];
			objs = [tr nodesForXPath:@"./recording/artist-credit/name-credit/artist/name" error:nil];
			if([objs count] == 1) [track setObject:[[objs objectAtIndex:0] stringValue] forKey:@"Artist"];
			else if([objs count]) {
				NSArray *credits = [tr nodesForXPath:@"./recording/artist-credit/name-credit" error:nil];
				NSMutableString *str = [NSMutableString string];
				int k;
				for(k=0;k<[credits count];k++) {
					id node = [credits objectAtIndex:k];
					NSString *artist = nil;
					NSString *joinphrase = nil;
					objs = [node nodesForXPath:@"./@joinphrase" error:nil];
					if([objs count]) joinphrase = [[objs objectAtIndex:0] stringValue];
					objs = [node nodesForXPath:@"./artist/name" error:nil];
					if([objs count]) artist = [[objs objectAtIndex:0] stringValue];
					if(artist && joinphrase) [str appendFormat:@"%@%@",artist,joinphrase];
					else if(artist) [str appendFormat:@"%@",artist];
				}
				[track setObject:str forKey:@"Artist"];
			}
			else {
				NSString *aartist = [release objectForKey:@"Artist"];
				if(aartist) [track setObject:aartist forKey:@"Artist"];
			}
			int trackNum = 0;
			objs = [tr nodesForXPath:@"./position" error:nil];
			if([objs count]) trackNum = [[[objs objectAtIndex:0] stringValue] intValue];
			if(trackNum) [trackList setObject:track forKey:[NSNumber numberWithInt:trackNum]];
		}
		[release setObject:trackList forKey:@"Tracks"];
	}
	//NSLog(@"%@",[release description]);
	return self;
}

- (void)dealloc
{
	[release release];
	[super dealloc];
}

- (NSDictionary *)disc
{
	return release;
}

@end
